import logging
import re
from difflib import ndiff
from typing import List

__all__ = ["remove_quotes", "string_diff"]


logger = logging.getLogger(__name__)


def remove_quotes(string: str):
    """Remove quotes (single or double quotes) from the string object passed in.

    Not only leading and trailing quotes are removed; also any quotes inside a string, if so, are removed.
    """
    # Pattern 1 removes only leading and trailing quotes. Quotes inside a string are kept.
    # @NOTE: Left here for documentation purposes only. Pattern 2 is the one used.
    # search_pattern = re.compile(r'(^[\'\\"]{1}|[\'\\"]{1}$)')

    # Pattern 2 removes ALL quotes in a string.
    # Meaning, not only leading and trailing quotes, but also quotes inside a string are removed.
    search_pattern = re.compile(r"[\'\"]")
    return re.sub(search_pattern, "", string)


def string_diff(text_1: str, text_2: str):
    """
    diff line by line
    printing only diff.
    """
    lines_1: List[str] = re.split("[\r\n]+", text_1)
    lines_2: List[str] = re.split("[\r\n]+", text_2)
    diffs: List[str] = []
    for index, item in enumerate(line for line in ndiff(lines_1, lines_2) if not re.search(r"^\s*$", line)):
        if re.match(r"^[+\-]", item):
            print("diff in line %4i:" % index, item)
            diffs.append((item))

    line_length = len(diffs)
    print(
        f"There {'is' if line_length == 1 else 'are'} {'no' if line_length == 0 else str(line_length)} difference{'' if line_length == 1 else 's'} between text 1 and text 2."
    )



from dictIO.cppDict import (  # noqa: F401
    CppDict as CppDict,
    find_global_key as find_global_key,
    global_key_exists as global_key_exists,
    order_keys as order_keys,
    set_global_key as set_global_key,
)
from dictIO.formatter import (  # noqa: F401
    Formatter as Formatter,
    CppFormatter as CppFormatter,
    FoamFormatter as FoamFormatter,
    JsonFormatter as JsonFormatter,
    XmlFormatter as XmlFormatter,
)
from dictIO.parser import (  # noqa: F401
    Parser as Parser,
    CppParser as CppParser,
    FoamParser as FoamParser,
    JsonParser as JsonParser,
    XmlParser as XmlParser,
)

from dictIO.dictReader import (  # noqa: F401
    DictReader as DictReader,
)
from dictIO.dictWriter import (  # noqa: F401
    DictWriter as DictWriter,
    create_target_file_name as create_target_file_name,
)
from dictIO.dictParser import (  # noqa: F401
    DictParser as DictParser,
)



from copy import deepcopy
from pathlib import Path

import pytest

from dictIO import (
    CppDict,
    CppParser,
    DictParser,
    DictReader,
    DictWriter,
    find_global_key,
    order_keys,
    set_global_key,
)


@pytest.fixture()
def test_dict():
    parser = CppParser()
    return parser.parse_file(Path("test_dict_dict"))


def test_init():  # sourcery skip: avoid-builtin-shadow
    dict = CppDict()
    assert dict.source_file is None
    assert dict.path == Path.cwd()
    assert dict.name == ""
    assert dict.line_content == []
    assert dict.line_comments == {}
    assert dict.includes == {}
    assert dict.block_content == ""
    assert dict.block_comments == {}
    assert dict.string_literals == {}
    assert dict.expressions == {}
    # assert dict.delimiters == ['{','}','[',']','(',')','<','>',';',',']
    assert dict.delimiters == ["{", "}", "(", ")", "<", ">", ";", ","]


def test_init_with_file():  # sourcery skip: avoid-builtin-shadow
    dict = CppDict("someDict")
    assert dict.path == Path.cwd()
    assert dict.source_file == Path.cwd() / "someDict"


def test_find_global_key():
    # sourcery skip: no-loop-in-tests
    # Prepare
    str_in_1 = "PLACEHOLDER000001"
    str_in_2 = "PLACEHOLDER000002"
    str_in_3 = "PLACEHOLDER000003"
    not_a_str_1 = 1234
    not_a_str_2 = 1.23
    not_a_str_3 = False

    key_1 = "key_1"
    key_2 = "key_2"
    key_3 = "key_3"
    key_n_1 = "key_n_1"
    key_n_2 = "key_n_2"
    key_n_3 = "key_n_3"
    keyd = "keyd"
    keyl = "keyl"
    keydl = "keydl"
    keyld = "keyld"
    keydld = "keydld"
    keyldl = "keyldl"
    d_nested = {
        key_1: str_in_1,
        key_2: str_in_2,
        key_3: str_in_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
        key_n_3: not_a_str_3,
    }
    l_nested = [
        str_in_1,
        str_in_2,
        str_in_3,
        not_a_str_1,
        not_a_str_2,
        not_a_str_3,
    ]
    dl_nested = {keyl: deepcopy(l_nested)}
    ld_nested = [deepcopy(d_nested)]
    dld_nested = {keyld: deepcopy(ld_nested)}
    ldl_nested = [deepcopy(dl_nested)]

    # Construct a dictionary dict_in with single entries, nested dicts and nested lists
    dict_in = {
        key_1: str_in_1,
        key_2: str_in_2,
        key_3: str_in_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
        key_n_3: not_a_str_3,
        keyd: d_nested,
        keyl: l_nested,
        keydl: dl_nested,
        keyld: ld_nested,
        keydld: dld_nested,
        keyldl: ldl_nested,
    }

    # Assert structure of dict_in
    assert len(dict_in) == 12
    assert len(dict_in[keyd]) == 6
    assert len(dict_in[keyl]) == 6
    assert len(dict_in[keydl]) == 1
    assert len(dict_in[keydl][keyl]) == 6
    assert len(dict_in[keyld]) == 1
    assert len(dict_in[keyld][0]) == 6
    assert len(dict_in[keydld]) == 1
    assert len(dict_in[keydld][keyld]) == 1
    assert len(dict_in[keydld][keyld][0]) == 6
    assert len(dict_in[keyldl]) == 1
    assert len(dict_in[keyldl][0]) == 1
    assert len(dict_in[keyldl][0][keyl]) == 6

    # Assert content of dict_in
    assert dict_in[key_1] == str_in_1
    assert dict_in[key_2] == str_in_2
    assert dict_in[key_3] == str_in_3
    assert dict_in[keyd][key_1] == str_in_1
    assert dict_in[keyd][key_2] == str_in_2
    assert dict_in[keyd][key_3] == str_in_3
    assert dict_in[keyl][0] == str_in_1
    assert dict_in[keyl][1] == str_in_2
    assert dict_in[keyl][2] == str_in_3
    assert dict_in[keydl][keyl][0] == str_in_1
    assert dict_in[keydl][keyl][1] == str_in_2
    assert dict_in[keydl][keyl][2] == str_in_3
    assert dict_in[keyld][0][key_1] == str_in_1
    assert dict_in[keyld][0][key_2] == str_in_2
    assert dict_in[keyld][0][key_3] == str_in_3
    assert dict_in[keydld][keyld][0][key_1] == str_in_1
    assert dict_in[keydld][keyld][0][key_2] == str_in_2
    assert dict_in[keydld][keyld][0][key_3] == str_in_3
    assert dict_in[keyldl][0][keyl][0] == str_in_1
    assert dict_in[keyldl][0][keyl][1] == str_in_2
    assert dict_in[keyldl][0][keyl][2] == str_in_3

    # Execute find_global_key and set_global_key
    dict_out = deepcopy(dict_in)
    str_out_1 = "string 1"
    str_out_2 = "string 2"
    str_out_3 = "string 3"
    global_key = find_global_key(arg=dict_out, query=str_in_1)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_1 with str_out_1
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_1)
        global_key = find_global_key(arg=dict_out, query=str_in_1)
    assert i == 7
    global_key = find_global_key(arg=dict_out, query=str_in_2)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_2 with str_out_2
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_2)
        global_key = find_global_key(arg=dict_out, query=str_in_2)
    assert i == 7
    global_key = find_global_key(arg=dict_out, query=str_in_3)
    i = 0
    while global_key:
        i += 1
        # Substitute str_in_3 with str_out_3
        set_global_key(arg=dict_out, global_key=global_key, value=str_out_3)
        global_key = find_global_key(arg=dict_out, query=str_in_3)
    assert i == 7

    # Assert structure of dict_out
    assert len(dict_out) == 12
    assert len(dict_out[keyd]) == 6
    assert len(dict_out[keyl]) == 6
    assert len(dict_out[keydl]) == 1
    assert len(dict_out[keydl][keyl]) == 6
    assert len(dict_out[keyld]) == 1
    assert len(dict_out[keyld][0]) == 6
    assert len(dict_out[keydld]) == 1
    assert len(dict_out[keydld][keyld]) == 1
    assert len(dict_out[keydld][keyld][0]) == 6
    assert len(dict_out[keyldl]) == 1
    assert len(dict_out[keyldl][0]) == 1
    assert len(dict_out[keyldl][0][keyl]) == 6

    # Assert content of dict_out
    assert dict_out[key_1] == str_out_1
    assert dict_out[key_2] == str_out_2
    assert dict_out[key_3] == str_out_3
    assert dict_out[keyd][key_1] == str_out_1
    assert dict_out[keyd][key_2] == str_out_2
    assert dict_out[keyd][key_3] == str_out_3
    assert dict_out[keyl][0] == str_out_1
    assert dict_out[keyl][1] == str_out_2
    assert dict_out[keyl][2] == str_out_3
    assert dict_out[keydl][keyl][0] == str_out_1
    assert dict_out[keydl][keyl][1] == str_out_2
    assert dict_out[keydl][keyl][2] == str_out_3
    assert dict_out[keyld][0][key_1] == str_out_1
    assert dict_out[keyld][0][key_2] == str_out_2
    assert dict_out[keyld][0][key_3] == str_out_3
    assert dict_out[keydld][keyld][0][key_1] == str_out_1
    assert dict_out[keydld][keyld][0][key_2] == str_out_2
    assert dict_out[keydld][keyld][0][key_3] == str_out_3
    assert dict_out[keyldl][0][keyl][0] == str_out_1
    assert dict_out[keyldl][0][keyl][1] == str_out_2
    assert dict_out[keyldl][0][keyl][2] == str_out_3


def test_order_keys():  # sourcery skip: avoid-builtin-shadow
    # sourcery skip: no-loop-in-tests
    # Prepare
    str_1 = "string 1"
    str_2 = "string 2"
    str_3 = "string 3"
    not_a_str_1 = 1234
    not_a_str_2 = 1.23
    not_a_str_3 = False

    key_1 = "key_1"
    key_2 = "key_2"
    key_3 = "key_3"
    key_n_1 = "key_n_1"
    key_n_2 = "key_n_2"
    key_n_3 = "key_n_3"
    key_4 = "key_4"

    d_nested = {
        key_3: str_3,
        key_1: str_1,
        key_2: str_2,
        key_n_3: not_a_str_3,
        key_n_1: not_a_str_1,
        key_n_2: not_a_str_2,
    }
    dict_in = {
        key_3: str_3,
        key_1: str_1,
        key_2: str_2,
        key_n_3: not_a_str_3,
        key_n_2: not_a_str_2,
        key_4: d_nested,
        key_n_1: not_a_str_1,
    }
    keys_expected = [key_1, key_2, key_3, key_4, key_n_1, key_n_2, key_n_3]
    keys_expected_nested = [key_1, key_2, key_3, key_n_1, key_n_2, key_n_3]

    dict = CppDict()
    dict.data.update(deepcopy(dict_in))

    # 1. negative test: assert dict_in is not alphanumerically ordered
    for index, key in enumerate(dict_in):
        assert key != keys_expected[index]
    for index, key in enumerate(dict_in[key_4]):
        assert key != keys_expected_nested[index]

    # 2. negative test: assert dict is not alphanumerically ordered
    for index, key in enumerate(dict.data):
        assert key != keys_expected[index]
    for index, key in enumerate(dict.data[key_4]):
        assert key != keys_expected_nested[index]

    dict_out = order_keys(dict_in)  # order_keys function defined in dict.py module
    dict.order_keys()  # order_keys instance method of CppDict class

    # 1. positive test for dict_out: assert dict_out is alphanumerically ordered
    for index, key in enumerate(dict_out):
        assert key == keys_expected[index]
    for index, key in enumerate(dict_out[key_4]):
        assert key == keys_expected_nested[index]

    # 2. positive test for dict: assert dict.data is alphanumerically ordered
    for index, key in enumerate(dict.data):
        assert key == keys_expected[index]
    for index, key in enumerate(dict.data[key_4]):
        assert key == keys_expected_nested[index]


def test_order_keys_of_test_dict(test_dict: CppDict):
    # Prepare
    # Execute
    test_dict.order_keys()
    # Assert
    assert str(test_dict.data["unordered"]) == str(test_dict.data["ordered"])


def test_reduce_scope_of_test_dict(test_dict: CppDict):
    # Prepare
    scope = ["scope", "subscope1"]
    # Execute
    test_dict.reduce_scope(scope)
    # Assert
    dict_out = test_dict.data
    assert len(dict_out) == 2  # subscope11, subscope12
    assert dict_out["subscope11"]["name"] == "subscope11"
    assert dict_out["subscope12"]["name"] == "subscope12"


def test_include():
    # Prepare
    source_dict_file = Path("test_dict_add_include")
    param_dict_file = Path("test_dict_paramDict")
    temp_dict_file = Path(f"{source_dict_file.name}_temp")
    parsed_dict_file = Path(f"parsed.{temp_dict_file.name}")
    temp_dict_file.unlink(missing_ok=True)
    parsed_dict_file.unlink(missing_ok=True)
    source_dict = DictReader.read(source_dict_file)
    param_dict = DictReader.read(param_dict_file)
    # Execute
    source_dict.include(param_dict)
    # Assert
    DictWriter.write(source_dict, temp_dict_file)
    parsed_dict = DictParser.parse(temp_dict_file)
    assert parsed_dict is not None
    assert parsed_dict["valueA"] == 1
    # Clean up
    temp_dict_file.unlink(missing_ok=True)
    parsed_dict_file.unlink(missing_ok=True)



from typing import Dict, List

__all__ = ["BorgCounter", "DejaVue"]


class BorgCounter:
    """
    A class that implements a static global counter.
    Instances of this class all share the same global counter.
    This is used in DictReader class to assure that multiple instances
    of CppDict do not generate conflicting IDs for placeholder strings
    (as would be the case otherwise when merging included dicts).
    """

    Borg: Dict[str, int] = {"theCount": -1}

    def __init__(self):
        self.__dict__ = BorgCounter.Borg

    def __call__(self) -> int:
        """Increments BorgCounter and returns next value.

        Returns
        -------
        int
            the next value (after incrementation)
        """
        self.theCount += 1
        #   Small tweak for our use case: As we have only six digits available in our placeholder strings,
        #   we don't want the counter to exceed this. Fair to start again at 0 then :-)
        if self.theCount > 999999:
            self.theCount = 0  # pyright: ignore
        return self.theCount

    @staticmethod
    def reset():
        """Reset the BorgCounter."""
        BorgCounter.Borg["theCount"] = -1


class DejaVue:
    """A class that implements a static global registry of strings."""

    djv: Dict[str, List[str]] = {"strings": []}

    def __init__(self):
        self.__dict__ = DejaVue.djv
        self.ret_val: bool = False

    def __call__(self, string: str) -> bool:
        """Call DejaVue with passed string.

        Checks whether the passed in string is already registered in DejaVue.
        If so, returns True. Otherwise False.

        Parameters
        ----------
        string : str
            the string to be checked

        Returns
        -------
        bool
            True if string is already registered. False otherwise.
        """

        if string in self.djv["strings"]:
            self.ret_val = True

        self.djv["strings"].append(string)

        return self.ret_val

    @property
    def strings(self) -> List[str]:
        """Return a list with all strings currently registered with DejaVue.

        Returns
        -------
        List[str]
            list with all strings currently registered with DejaVue.
        """
        return self.djv["strings"]

    def reset(self):
        """Reset DejaVue.

        Clears the list of registered strings.
        """
        self.djv["strings"] = []
        self.ret_val = False
